/*
 * Minio Javascript Library for Amazon S3 Compatible Cloud Storage, (C) 2016 Minio, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

require('source-map-support').install()

import { assert } from 'chai';
import Concat from 'concat-stream';
import Http from 'http';
import Nock from 'nock';
import Through2 from 'through2';
import Stream from 'stream';
import Minio from '../../../dist/main/minio.js';
import MockTransport from './transport.js';

var Package = require('../../../package.json')

describe('Client', () => {
  var nockRequests = []
  beforeEach(() => {
    Nock.cleanAll()
    nockRequests = []
  })
  afterEach(() => {
    nockRequests.forEach(element => {
      if (!element.request.isDone()) {
        element.request.done()
      }
    })
  })
  var MockResponse = (address) => {
      var request = Nock(address),
        trace = new Error().stack
      nockRequests.push({
        request: request,
        trace: trace
      })
      return request
    },
    client = new Minio({
      url: 'http://localhost:9000',
      accessKey: 'accesskey',
      secretKey: 'secretkey'
    })
  describe('new client', () => {
    it('should work with http', () => {
      var client = new Minio({
        url: 'http://localhost',
        accessKey: 'accesskey',
        secretKey: 'secretkey'
      })
      assert.equal(client.params.port, 80)
    })
    it('should override port with http', () => {
      var client = new Minio({
        url: 'http://localhost:9000',
        accessKey: 'accesskey',
        secretKey: 'secretkey'
      })
      assert.equal(client.params.port, 9000)
    })
    it('should work with https', () => {
      var client = new Minio({
        url: 'https://localhost',
        accessKey: 'accesskey',
        secretKey: 'secretkey'
      })
      assert.equal(client.params.port, 443)
    })
    it('should override port with https', () => {
      var client = new Minio({
        url: 'https://localhost:9000',
        accessKey: 'accesskey',
        secretKey: 'secretkey'
      })
      assert.equal(client.params.port, 9000)
    })
    it('should fail with no url', (done) => {
      try {
        new Minio({
          accessKey: 'accesskey',
          secretKey: 'secretkey'
        })
      } catch (e) {
        done()
      }
    })
    it('should fail with no scheme', (done) => {
      try {
        new Minio({
          url: 'localhost',
          accessKey: 'accesskey',
          secretKey: 'secretkey'
        })
      } catch (e) {
        done()
      }
    })
  })
  describe('Presigned URL', () => {
    describe('presigned-get', () => {
      it('should not generate presigned url with no access key', (done) => {
        try {
          var client = new Minio({
            url: 'https://localhost:9000',
          })
          client.presignedGetObject('bucket', 'object', '1000')
        } catch (e) {
          done()
        }
      })
      it('should not generate presigned url with wrong expires param', (done) => {
        try {
          var client = new Minio({
            url: 'https://localhost:9000',
            accessKey: 'accesskey',
            secretKey: 'secretkey',
          })
          client.presignedGetObject('bucket', 'object', '0')
        } catch (e) {
          done()
        }
      })
      it('should generate presigned url', () => {
        var client = new Minio({
          url: 'https://localhost:9000',
          accessKey: 'accesskey',
          secretKey: 'secretkey'
        })
        var url = client.presignedGetObject('bucket', 'object', '86400')
        assert.equal(url.length > 0, true)
      })
    })
    describe('presigned-put', () => {
      it('should not generate presigned url with no access key', (done) => {
        try {
          var client = new Minio({
            url: 'https://localhost:9000',
          })
          client.presignedPutObject('bucket', 'object', '1000')
        } catch (e) {
          done()
        }
      })
      it('should not generate presigned url with wrong expires param', (done) => {
        try {
          var client = new Minio({
            url: 'https://localhost:9000',
            accessKey: 'accesskey',
            secretKey: 'secretkey',
          })
          client.presignedPutObject('bucket', 'object', '0')
        } catch (e) {
          done()
        }
      })
      it('should generate presigned url', () => {
        var client = new Minio({
          url: 'https://localhost:9000',
          accessKey: 'accesskey',
          secretKey: 'secretkey'
        })
        var url = client.presignedPutObject('bucket', 'object', '86400')
        assert.equal(url.length > 0, true)
      })
    })
  })
  describe('User Agent', () => {
    it('should have a default user agent', () => {
      var client = new Minio({
        url: 'https://localhost:9000',
        accessKey: 'accesskey',
        secretKey: 'secretkey'
      })
      assert.equal(`minio-js/${Package.version} (${process.platform}; ${process.arch})`, client.params.userAgent)
    })
    it('should set user agent', () => {
      var client = new Minio({
        url: 'https://localhost:9000',
        accessKey: 'accesskey',
        secretKey: 'secretkey'
      })
      client.setUserAgent('test', '1.0.0', ['comment', 'arch', 'os'])
      assert.equal(`minio-js/${Package.version} (${process.platform}; ${process.arch}) test/1.0.0 (comment; arch; os)`, client.params.userAgent)
    })
    it('should set user agent without comments', () => {
      var client = new Minio({
        url: 'https://localhost:9000',
        accessKey: 'accesskey',
        secretKey: 'secretkey'
      })
      client.setUserAgent('test', '1.0.0', [])
      assert.equal(`minio-js/${Package.version} (${process.platform}; ${process.arch}) test/1.0.0`, client.params.userAgent)
    })
    it('should not set user agent without name', (done) => {
      try {
        var client = new Minio({
          url: 'https://localhost:9000',
          accessKey: 'accesskey',
          secretKey: 'secretkey'
        })
        client.setUserAgent(null, '1.0.0')
      } catch (e) {
        done()
      }
    })
    it('should not set user agent with empty name', (done) => {
      try {
        var client = new Minio({
          url: 'https://localhost:9000',
          accessKey: 'accesskey',
          secretKey: 'secretkey'
        })
        client.setUserAgent('', '1.0.0')
      } catch (e) {
        done()
      }
    })
    it('should not set user agent without version', (done) => {
      try {
        var client = new Minio({
          url: 'https://localhost:9000',
          accessKey: 'accesskey',
          secretKey: 'secretkey'
        })
        client.setUserAgent('test', null)
      } catch (e) {
        done()
      }
    })
    it('should not set user agent with empty version', (done) => {
      try {
        var client = new Minio({
          url: 'https://localhost:9000',
          accessKey: 'accesskey',
          secretKey: 'secretkey'
        })
        client.setUserAgent('test', '')
      } catch (e) {
        done()
      }
    })
    it('should not set user agent twice', (done) => {
      try {
        var client = new Minio({
          url: 'https://localhost:9000',
          accessKey: 'accesskey',
          secretKey: 'secretkey'
        })
        client.setUserAgent('test', '1.0.0')
        assert.equal(`minio-js/${Package.version} (${process.platform}; ${process.arch}) test/1.0.0`, client.params.userAgent)
        client.setUserAgent('test', '1.0.0')
      } catch (e) {
        done()
      }
    })
  })
  describe('Authentication', () => {
    describe('not set', () => {
      var transport = new MockTransport(),
        client = new Minio({
          url: 'http://localhost:9000'
        }, transport)
      it('should not send auth info without keys', (done) => {
        client.transport.addRequest((params) => {
          assert.deepEqual(params, {
            host: 'localhost',
            port: 9000,
            protocol: '',
            path: '/bucket/object',
            method: 'HEAD'
          })
        }, 200, {
          'etag': 'etag',
          'content-length': 11,
          'last-modified': 'lastmodified',
          'content-type': 'application/octet-stream'
        }, null)
        client.statObject('bucket', 'object', (e, r) => {
          assert.deepEqual(r, {
            size: 11,
            'lastModified': 'lastmodified',
            etag: 'etag',
            contentType: 'application/octet-stream'
          })
          done()
        })
      })
    })
    describe('set with access and secret keys', () => {
      it('should not send auth info without keys', (done) => {
        var transport = new MockTransport(),
          client = new Minio({
            url: 'http://localhost:9000',
            accessKey: 'accessKey',
            secretKey: 'secretKey'
          }, transport)
        client.transport.addRequest((params) => {
          assert.equal(true, params.headers.authorization !== null)
          assert.equal(true, params.headers.authorization.indexOf('accessKey') > -1)
          assert.equal(true, params.headers['x-amz-date'] !== null)
          delete params.headers.authorization
          delete params.headers['x-amz-date']
          assert.deepEqual(params, {
            host: 'localhost',
            port: 9000,
            protocol: '',
            path: '/bucket/object',
            method: 'HEAD',
            headers: {
              host: 'localhost',
              'x-amz-content-sha256': 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
            }
          })
        }, 200, {
          'etag': 'etag',
          'content-length': 11,
          'last-modified': 'lastmodified',
          'content-type': 'text/plain'
        }, null)
        client.statObject('bucket', 'object', (e, r) => {
          assert.deepEqual(r, {
            size: 11,
            'lastModified': 'lastmodified',
            etag: 'etag',
            contentType: 'text/plain'
          })
          done()
        })
      })
    })
  })

  describe('Bucket API calls', () => {
    describe('#makeBucket(bucket, callback)', () => {
      it('should call the callback on success', (done) => {
        MockResponse('http://localhost:9000').put('/bucket').reply(200)
        client.makeBucket('bucket', done)
      })
      it('pass an error into the callback on failure', (done) => {
        MockResponse('http://localhost:9000').put('/bucket').reply(400, generateError('code', 'message', 'requestid', 'hostid', '/bucket'))
        client.makeBucket('bucket', checkError('code', 'message', 'requestid', 'hostid', '/bucket', done))
      })
      it('should fail on null bucket', (done) => {
        try {
          client.makeBucket(null)
        } catch (e) {
          done()
        }
      })
      it('should fail on empty bucket', (done) => {
        try {
          client.makeBucket('')
        } catch (e) {
          done()
        }
      })
      it('should fail on empty bucket', (done) => {
        try {
          client.makeBucket('  \n  \t  ')
        } catch (e) {
          done()
        }
      })
    })

    describe('#listBuckets()', () => {
      it('should generate a bucket iterator', (done) => {
        MockResponse('http://localhost:9000').get('/').reply(200, '<ListAllMyBucketsResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner><Buckets><Bucket><Name>bucket</Name><CreationDate>2015-05-05T20:35:51.410Z</CreationDate></Bucket><Bucket><Name>foo</Name><CreationDate>2015-05-05T20:35:47.170Z</CreationDate></Bucket></Buckets></ListAllMyBucketsResult>')
        var results = []
        var expectedResults = [{
          name: 'bucket',
          creationDate: '2015-05-05T20:35:51.410Z'
        }, {
          name: 'foo',
          creationDate: '2015-05-05T20:35:47.170Z'
        }]
        client.listBuckets(function(e, stream) {
          stream.on('data', function(obj) {
            results.push(obj)
          })
          stream.on('end', function() {
            assert.deepEqual(results, expectedResults)
            done()
          })
        })
      })
      it('should pass error to callback', (done) => {
        MockResponse('http://localhost:9000').get('/').reply(400, generateError('code', 'message', 'requestid', 'hostid', '/'))
        client.listBuckets(function(e, stream) {
          checkError('code', 'message', 'requestid', 'hostid', '/')(e)
          done()
        })
      })
      it('should convert 307 to 403', (done) => {
        MockResponse('http://localhost:9000').get('/').reply(307)
        client.listBuckets(function(e, stream) {
          checkError('AccessDenied', 'Valid and authorized credentials required', null, null, null)(e)
          done()
        })
      })
    })

    describe('#bucketExists(bucket, cb)', () => {
      it('should call callback with no options if successful', (done) => {
        MockResponse('http://localhost:9000').head('/bucket').reply(204)
        client.bucketExists('bucket', (e) => {
          assert.equal(e, null)
          done()
        })
      })
      it('should pass error to callback', (done) => {
        MockResponse('http://localhost:9000').head('/bucket').reply(400, generateError('code', 'message', 'requestid', 'hostid', 'resource'))
        client.bucketExists('bucket', checkError('code', 'message', 'requestid', 'hostid', 'resource', (r) => {
          assert.equal(r, null)
          done()
        }))
      })
      it('should return an error on moved permanently', (done) => {
        MockResponse('http://localhost:9000').head('/bucket').reply(301)
        client.bucketExists('bucket', checkError('MovedPermanently', 'Moved Permanently', null, null, null, (r) => {
          assert.equal(r, null)
          done()
        }))
      })
      it('should return an error on 404', (done) => {
        MockResponse('http://localhost:9000').head('/bucket').reply(403)
        client.bucketExists('bucket', checkError('AccessDenied', 'Valid and authorized credentials required', null, null, null, (r) => {
          assert.equal(r, null)
          done()
        }))
      })
      it('should return an error on 404', (done) => {
        MockResponse('http://localhost:9000').head('/bucket').reply(404)
        client.bucketExists('bucket', checkError('NotFound', 'Not Found', null, null, null, (r) => {
          assert.equal(r, null)
          done()
        }))
      })
      it('should fail on null bucket', (done) => {
        try {
          client.bucketExists(null)
        } catch (e) {
          done()
        }
      })
      it('should fail on empty bucket', (done) => {
        try {
          client.bucketExists('')
        } catch (e) {
          done()
        }
      })
      it('should fail on empty bucket', (done) => {
        try {
          client.BucketExists('  \n  \t  ')
        } catch (e) {
          done()
        }
      })
    })

    describe('#removeBucket(bucket, cb)', () => {
      it('should remove a bucket', (done) => {
        MockResponse('http://localhost:9000').delete('/bucket').reply(204)
        client.removeBucket('bucket', () => {
          done()
        })
      })
      it('should pass error to callback', (done) => {
        MockResponse('http://localhost:9000').delete('/bucket').reply(400, generateError('code', 'message', 'requestid', 'hostid', 'resource'))
        client.removeBucket('bucket', checkError('code', 'message', 'requestid', 'hostid', 'resource', (r) => {
          assert.equal(r, null)
          done()
        }))
      })
      it('should fail on null bucket', (done) => {
        try {
          client.removeBucket(null)
        } catch (e) {
          done()
        }
      })
      it('should fail on empty bucket', (done) => {
        try {
          client.removeBucket('')
        } catch (e) {
          done()
        }
      })
      it('should fail on empty bucket', (done) => {
        try {
          client.removeBucket('  \n  \t  ')
        } catch (e) {
          done()
        }
      })
    })

    describe('#getBucketACL(bucket, cb)', () => {
      it('should return public-read-write acl', (done) => {
        MockResponse('http://localhost:9000').get('/bucket?acl').reply(200, '<AccessControlPolicy xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Owner><ID>75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a</ID><DisplayName>CustomersName@amazon.com</DisplayName></Owner><AccessControlList><Grant><Grantee xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:type="CanonicalUser"><ID>75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a</ID><DisplayName>CustomersName@amazon.com</DisplayName><URI>http://acs.amazonaws.com/groups/global/AllUsers</URI></Grantee><Permission>WRITE</Permission></Grant><Grant><Grantee xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:type="CanonicalUser"><ID>75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a</ID><DisplayName>CustomersName@amazon.com</DisplayName><URI>http://acs.amazonaws.com/groups/global/AllUsers</URI></Grantee><Permission>READ</Permission></Grant><Grant><Grantee xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:type="CanonicalUser"><ID>75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a</ID><DisplayName>CustomersName@amazon.com</DisplayName></Grantee><Permission>FULL_CONTROL</Permission></Grant></AccessControlList></AccessControlPolicy>')
        client.getBucketACL('bucket', (e, r) => {
          assert.equal(e, null)
          assert.equal(r, 'public-read-write')
          done()
        })
      })
      it('should return public-read acl', (done) => {
        MockResponse('http://localhost:9000').get('/bucket?acl').reply(200, '<AccessControlPolicy xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Owner><ID>75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a</ID><DisplayName>CustomersName@amazon.com</DisplayName></Owner><AccessControlList><Grant><Grantee xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:type="CanonicalUser"><ID>75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a</ID><DisplayName>CustomersName@amazon.com</DisplayName><URI>http://acs.amazonaws.com/groups/global/AllUsers</URI></Grantee><Permission>READ</Permission></Grant><Grant><Grantee xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:type="CanonicalUser"><ID>75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a</ID><DisplayName>CustomersName@amazon.com</DisplayName></Grantee><Permission>FULL_CONTROL</Permission></Grant></AccessControlList></AccessControlPolicy>')
        client.getBucketACL('bucket', (e, r) => {
          assert.equal(e, null)
          assert.equal(r, 'public-read')
          done()
        })
      })
      it('should return authenticated-read acl', (done) => {
        MockResponse('http://localhost:9000').get('/bucket?acl').reply(200, '<AccessControlPolicy xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Owner><ID>75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a</ID><DisplayName>CustomersName@amazon.com</DisplayName></Owner><AccessControlList><Grant><Grantee xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:type="CanonicalUser"><ID>75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a</ID><DisplayName>CustomersName@amazon.com</DisplayName><URI>http://acs.amazonaws.com/groups/global/AuthenticatedUsers</URI></Grantee><Permission>READ</Permission></Grant><Grant><Grantee xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:type="CanonicalUser"><ID>75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a</ID><DisplayName>CustomersName@amazon.com</DisplayName></Grantee><Permission>FULL_CONTROL</Permission></Grant></AccessControlList></AccessControlPolicy>')
        client.getBucketACL('bucket', (e, r) => {
          assert.equal(e, null)
          assert.equal(r, 'authenticated-read')
          done()
        })
      })
      it('should return private acl', (done) => {
        MockResponse('http://localhost:9000').get('/bucket?acl').reply(200, '<AccessControlPolicy xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Owner><ID>75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a</ID><DisplayName>CustomersName@amazon.com</DisplayName></Owner><AccessControlList><Grant><Grantee xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:type="CanonicalUser"><ID>75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a</ID><DisplayName>CustomersName@amazon.com</DisplayName></Grantee><Permission>FULL_CONTROL</Permission></Grant></AccessControlList></AccessControlPolicy>')
        client.getBucketACL('bucket', (e, r) => {
          assert.equal(e, null)
          assert.equal(r, 'private')
          done()
        })
      })
      it('should pass error to callback', (done) => {
        MockResponse('http://localhost:9000').get('/bucket?acl').reply(400, generateError('code', 'message', 'requestid', 'hostid', 'resource'))
        client.getBucketACL('bucket', checkError('code', 'message', 'requestid', 'hostid', 'resource', done))
      })
      it('should fail on null bucket', (done) => {
        try {
          client.getBucketACL(null)
        } catch (e) {
          done()
        }
      })
      it('should fail on empty bucket', (done) => {
        try {
          client.getBucketACL('')
        } catch (e) {
          done()
        }
      })
      it('should fail on empty bucket', (done) => {
        try {
          client.getBucketACL('  \n  \t  ')
        } catch (e) {
          done()
        }
      })
    })

    describe('#setBucketACL(bucket, acl, cb)', () => {
      it('should set acl', (done) => {
        var transport = new MockTransport(),
          client = new Minio({
            url: 'http://localhost:9000'
          }, transport)
        client.transport.addRequest((params) => {
          assert.deepEqual(params, {
            host: 'localhost',
            port: 9000,
            path: '/bucket?acl',
            method: 'PUT',
            protocol: '',
            headers: {
              'x-amz-acl': 'public'
            }
          })
        }, 200, {
          'etag': 'etag',
          'content-length': 11,
          'last-modified': 'lastmodified'
        }, null)
        client.setBucketACL('bucket', 'public', (e) => {
          assert.equal(e, null)
        })
        done()
      })
      it('should pass error to callback', (done) => {
        MockResponse('http://localhost:9000').put('/bucket?acl').reply(400, generateError('code', 'message', 'requestid', 'hostid', 'resource'))
        client.setBucketACL('bucket', 'public', checkError('code', 'message', 'requestid', 'hostid', 'resource', (r) => {
          assert.equal(r, null)
          done()
        }))
      })
      it('should fail on null bucket', (done) => {
        try {
          client.setBucketACL(null, 'public')
        } catch (e) {
          done()
        }
      })
      it('should fail on empty bucket', (done) => {
        try {
          client.setBucketACL('', 'public')
        } catch (e) {
          done()
        }
      })
      it('should fail on empty bucket', (done) => {
        try {
          client.setBucketACL('  \n  \t  ', 'public')
        } catch (e) {
          done()
        }
      })
      it('should fail on null acl', (done) => {
        try {
          client.setBucketACL('hello', null)
        } catch (e) {
          done()
        }
      })
      it('should fail on empty acl', (done) => {
        try {
          client.setBucketACL('hello', '')
        } catch (e) {
          done()
        }
      })
      it('should fail on empty acl', (done) => {
        try {
          client.setBucketACL('hello', '  \n  \t  ')
        } catch (e) {
          done()
        }
      })
    })
  })

  describe('object level', () => {
    describe('#getObject(bucket, object, callback)', () => {
      it('should return a stream object', (done) => {
        MockResponse('http://localhost:9000').get('/bucket/object').reply(200, 'hello world')
        client.getObject('bucket', 'object', (e, r) => {
          assert.equal(e, null)
          r.pipe(Concat(buf => {
            assert.equal(buf, 'hello world')
            done()
          }))
        })
      })
      it('should pass error to callback', (done) => {
        MockResponse('http://localhost:9000').get('/bucket/object').reply(400, generateError('code', 'message', 'requestid', 'hostid', '/bucket/object'))
        client.getObject('bucket', 'object', checkError('code', 'message', 'requestid', 'hostid', '/bucket/object', (r) => {
          assert.equal(r, null)
          done()
        }))
      })
      it('should fail on null bucket', (done) => {
        try {
          client.getObject(null, 'hello')
        } catch (e) {
          done()
        }
      })
      it('should fail on empty bucket', (done) => {
        try {
          client.getObject('', 'hello')
        } catch (e) {
          done()
        }
      })
      it('should fail on empty bucket', (done) => {
        try {
          client.getObject('  \n  \t  ', 'hello')
        } catch (e) {
          done()
        }
      })
      it('should fail on null object', (done) => {
        try {
          client.getObject('hello', null)
        } catch (e) {
          done()
        }
      })
      it('should fail on empty object', (done) => {
        try {
          client.getObject('hello', '')
        } catch (e) {
          done()
        }
      })
      it('should fail on empty object', (done) => {
        try {
          client.getObject('hello', '  \n  \t  ')
        } catch (e) {
          done()
        }
      })
    })
    describe('#getPartialObject(bucket, object, offset, range) callback)', () => {
      it('should work with offset and range', (done) => {
        MockResponse('http://localhost:9000', {
          reqHeaders: {
            'range': '10-21'
          }
        }).get('/bucket/object').reply(206, 'hello world')
        client.getPartialObject('bucket', 'object', 10, 11, (e, r) => {
          assert.equal(e, null)
          r.pipe(Concat(buf => {
            assert.equal(buf, 'hello world')
            done()
          }))
        })
      })
      it('should work with offset', (done) => {
        MockResponse('http://localhost:9000', {
          reqHeaders: {
            'range': '10-'
          }
        }).get('/bucket/object').reply(206, 'hello world')
        client.getPartialObject('bucket', 'object', 10, null, (e, r) => {
          assert.equal(e, null)
          r.pipe(Concat(buf => {
            assert.equal(buf, 'hello world')
            done()
          }))
        })
      })
      it('should work with range', (done) => {
        MockResponse('http://localhost:9000', {
          reqHeaders: {
            'range': '0-21'
          }
        }).get('/bucket/object').reply(206, 'hello world')
        client.getPartialObject('bucket', 'object', null, 11, (e, r) => {
          assert.equal(e, null)
          r.pipe(Concat(buf => {
            assert.equal(buf, 'hello world')
            done()
          }))
        })
      })
    })

    describe('#putObject(bucket, object, contentType, size, source, callback)', () => {
      describe('with small objects using single put', () => {
        it('should put an object', (done) => {
          MockResponse('http://localhost:9000').put('/bucket/object', 'hello world').reply(200)
          var s = new Stream.Readable()
          s._read = function() {}
          s.push('hello world')
          s.push(null)
          client.putObject('bucket', 'object', '', 11, s, done)
        })
        it('should pass error to callback', (done) => {
          MockResponse('http://localhost:9000').put('/bucket/object', 'hello world').reply(400, generateError('code', 'message', 'requestid', 'hostid', 'resource'))
          var s = new Stream.Readable()
          s._read = function() {}
          s.push('hello world')
          s.push(null)
          client.putObject('bucket', 'object', '', 11, s, checkError('code', 'message', 'requestid', 'hostid', 'resource', done))
        })
        it('should fail when data is smaller than specified', (done) => {
          var s = new Stream.Readable()
          s._read = function() {}
          s.push('hello world')
          s.push(null)
          client.putObject('bucket', 'object', '', 12, s, (e) => {
            if (e) {
              done()
            }
          })
        })
        it('should fail when data is larger than specified', (done) => {
          var s = new Stream.Readable()
          s._read = function() {}
          s.push('hello world')
          s.push(null)
          client.putObject('bucket', 'object', '', 10, s, (e) => {
            if (e) {
              done()
            }
          })
        })
        it('should fail on null bucket', (done) => {
          try {
            client.putObject(null, 'hello', '', 1)
          } catch (e) {
            done()
          }
        })
        it('should fail on empty bucket', (done) => {
          try {
            client.putObject(' \n \t ', 'hello', '', 1)
          } catch (e) {
            done()
          }
        })
        it('should fail on empty bucket', (done) => {
          try {
            client.putObject('', 'hello', '', 1)
          } catch (e) {
            done()
          }
        })
        it('should fail on null object', (done) => {
          try {
            client.putObject('hello', null, '', 1)
          } catch (e) {
            done()
          }
        })
        it('should fail on empty object', (done) => {
          try {
            client.putObject('hello', '', '', 1)
          } catch (e) {
            done()
          }
        })
        it('should fail on empty object', (done) => {
          try {
            client.putObject('hello', ' \n \t ', '', 1)
          } catch (e) {
            done()
          }
        })
      })
      describe('with large objects using multipart', () => {
        var uploadData = ''
        for (var i = 0; i < 1024; i++) {
          uploadData += 'a'
        }
        it('should put an object with no resume needed', (done) => {
          MockResponse('http://localhost:9000').get('/bucket?uploads&max-uploads=1000&prefix=object').reply(200, '<ListMultipartUploadsResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Bucket>golang</Bucket><KeyMarker></KeyMarker><UploadIdMarker></UploadIdMarker><NextKeyMarker></NextKeyMarker><NextUploadIdMarker></NextUploadIdMarker><EncodingType></EncodingType><MaxUploads>1000</MaxUploads><IsTruncated>false</IsTruncated><Prefix></Prefix><Delimiter></Delimiter></ListMultipartUploadsResult>')
          MockResponse('http://localhost:9000').post('/bucket/object?uploads').reply(200, '<?xml version="1.0" encoding="UTF-8"?>\n<InitiateMultipartUploadResult><Bucket>bucket</Bucket><Key>object</Key><UploadId>uploadid</UploadId></InitiateMultipartUploadResult>')
          MockResponse('http://localhost:9000').put('/bucket/object?partNumber=1&uploadId=uploadid', (body) => {
            return body.length === 5 * 1024 * 1024

          }).reply(200, '', {
            etag: 'etag1'
          })
          MockResponse('http://localhost:9000').put('/bucket/object?partNumber=2&uploadId=uploadid', (body) => {
            return body.length === 5 * 1024 * 1024

          }).reply(200, '', {
            etag: 'etag2'
          })
          MockResponse('http://localhost:9000').put('/bucket/object?partNumber=3&uploadId=uploadid', (body) => {
            return body.length === 1024 * 1024

          }).reply(200, '', {
            etag: 'etag3'
          })
          MockResponse('http://localhost:9000').post('/bucket/object?uploadId=uploadid').reply(200, '<?xml version="1.0" encoding="UTF-8"?><CompleteMultipartUploadResult><Bucket>bucket</Bucket><Key>object</Key><Location>location</Location><ETag>"3858f62230ac3c915f300c664312c11f"</ETag></CompleteMultipartUploadResult>')
          var s = new Stream.Readable()
          s._read = function() {}
          for (var i = 0; i < 11 * 1024; i++) {
            s.push(uploadData)
          }
          s.push(null)
          client.putObject('bucket', 'object', '', 11 * 1024 * 1024, s, done)
        })
        it('should resume an object upload', (done) => {
          MockResponse('http://localhost:9000').get('/bucket?uploads&max-uploads=1000&prefix=object').reply(200, '<ListMultipartUploadsResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Bucket>bucket</Bucket><KeyMarker></KeyMarker><UploadIdMarker></UploadIdMarker><NextKeyMarker></NextKeyMarker><NextUploadIdMarker></NextUploadIdMarker><EncodingType></EncodingType><MaxUploads>1000</MaxUploads><IsTruncated>false</IsTruncated><Upload><Key>object</Key><UploadId>uploadid</UploadId><Initiator><ID></ID><DisplayName></DisplayName></Initiator><Owner><ID></ID><DisplayName></DisplayName></Owner><StorageClass></StorageClass><Initiated>2015-05-30T14:43:35.349Z</Initiated></Upload><Prefix>object</Prefix><Delimiter></Delimiter></ListMultipartUploadsResult>')
          MockResponse('http://localhost:9000').get('/bucket/object?uploadId=uploadid').reply(200, '<ListPartsResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Bucket>bucket</Bucket><Key>go1.4.2</Key><UploadId>ntWSjzBytPT2xKLaMRonzXncsO10EH4Fc-Iq2-4hG-ulRYB</UploadId><Initiator><ID>minio</ID><DisplayName>minio</DisplayName></Initiator><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner><StorageClass>STANDARD</StorageClass><PartNumberMarker>0</PartNumberMarker><NextPartNumberMarker>0</NextPartNumberMarker><MaxParts>1000</MaxParts><IsTruncated>false</IsTruncated><Part><PartNumber>1</PartNumber><ETag>"79b281060d337b9b2b84ccf390adcf74"</ETag><LastModified>2015-06-03T03:12:34.756Z</LastModified><Size>5242880</Size></Part><Part><PartNumber>2</PartNumber><ETag>"79b281060d337b9b2b84ccf390adcf74"</ETag><LastModified>2015-06-03T03:12:34.756Z</LastModified><Size>5242880</Size></Part></ListPartsResult>')
          MockResponse('http://localhost:9000').put('/bucket/object?partNumber=3&uploadId=uploadid', (body) => {
            return body.length === 1024 * 1024

          }).reply(200, '', {
            etag: '79b281060d337b9b2b84ccf390adcf74'
          })
          MockResponse('http://localhost:9000').post('/bucket/object?uploadId=uploadid').reply(200, '<?xml version="1.0" encoding="UTF-8"?><CompleteMultipartUploadResult><Bucket>bucket</Bucket><Key>object</Key><Location>location</Location><ETag>"3858f62230ac3c915f300c664312c11f"</ETag></CompleteMultipartUploadResult>')
          var s = new Stream.Readable()
          s._read = function() {}
          for (var i = 0; i < 11 * 1024; i++) {
            s.push(uploadData)
          }
          s.push(null)
          client.putObject('bucket', 'object', '', 11 * 1024 * 1024, s, (e) => {
            done(e)
          })
        })
        it('should resume an object upload when uploaded data does not match, overwriting mismatching parts', (done) => {
          MockResponse('http://localhost:9000').get('/bucket?uploads&max-uploads=1000&prefix=object').reply(200, '<ListMultipartUploadsResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Bucket>bucket</Bucket><KeyMarker></KeyMarker><UploadIdMarker></UploadIdMarker><NextKeyMarker></NextKeyMarker><NextUploadIdMarker></NextUploadIdMarker><EncodingType></EncodingType><MaxUploads>1000</MaxUploads><IsTruncated>false</IsTruncated><Upload><Key>object</Key><UploadId>uploadid</UploadId><Initiator><ID></ID><DisplayName></DisplayName></Initiator><Owner><ID></ID><DisplayName></DisplayName></Owner><StorageClass></StorageClass><Initiated>2015-05-30T14:43:35.349Z</Initiated></Upload><Prefix>object</Prefix><Delimiter></Delimiter></ListMultipartUploadsResult>')
          MockResponse('http://localhost:9000').get('/bucket/object?uploadId=uploadid').reply(200, '<ListPartsResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Bucket>bucket</Bucket><Key>go1.4.2</Key><UploadId>ntWSjzBytPT2xKLaMRonzXncsO10EH4Fc-Iq2-4hG-ulRYB</UploadId><Initiator><ID>minio</ID><DisplayName>minio</DisplayName></Initiator><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner><StorageClass>STANDARD</StorageClass><PartNumberMarker>0</PartNumberMarker><NextPartNumberMarker>0</NextPartNumberMarker><MaxParts>1000</MaxParts><IsTruncated>false</IsTruncated><Part><PartNumber>1</PartNumber><ETag>"79b281060d337b9b2b84ccf390adcf74"</ETag><LastModified>2015-06-03T03:12:34.756Z</LastModified><Size>5242880</Size></Part><Part><PartNumber>2</PartNumber><ETag>"89b281060d337b9b2b84ccf390adcf74"</ETag><LastModified>2015-06-03T03:12:34.756Z</LastModified><Size>5242880</Size></Part></ListPartsResult>')
          MockResponse('http://localhost:9000').put('/bucket/object?partNumber=2&uploadId=uploadid', (body) => {
            return body.length === 5 * 1024 * 1024
          }).reply(200, '', {
            etag: '79b281060d337b9b2b84ccf390adcf74'
          })
          MockResponse('http://localhost:9000').put('/bucket/object?partNumber=3&uploadId=uploadid', (body) => {
            return body.length === 1024 * 1024

          }).reply(200, '', {
            etag: '79b281060d337b9b2b84ccf390adcf74'
          })
          MockResponse('http://localhost:9000').post('/bucket/object?uploadId=uploadid').reply(200, '<?xml version="1.0" encoding="UTF-8"?><CompleteMultipartUploadResult><Bucket>bucket</Bucket><Key>object</Key><Location>location</Location><ETag>"3858f62230ac3c915f300c664312c11f"</ETag></CompleteMultipartUploadResult>')
          var s = new Stream.Readable()
          s._read = function() {}
          for (var i = 0; i < 11 * 1024; i++) {
            s.push(uploadData)
          }
          s.push(null)
          client.putObject('bucket', 'object', '', 11 * 1024 * 1024, s, done)
        })
        it('should fail if actual size is smaller than expected', (done) => {
          MockResponse('http://localhost:9000').get('/bucket?uploads&max-uploads=1000&prefix=object').reply(200, '<ListMultipartUploadsResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Bucket>golang</Bucket><KeyMarker></KeyMarker><UploadIdMarker></UploadIdMarker><NextKeyMarker></NextKeyMarker><NextUploadIdMarker></NextUploadIdMarker><EncodingType></EncodingType><MaxUploads>1000</MaxUploads><IsTruncated>false</IsTruncataed><Prefix></Prefix><Delimiter></Delimiter></ListMultipartUploadsResult>')
          MockResponse('http://localhost:9000').post('/bucket/object?uploads').reply(200, '<?xml version="1.0" encoding="UTF-8"?>\n<InitiateMultipartUploadResult><Bucket>bucket</Bucket><Key>object</Key><UploadId>uploadid</UploadId></InitiateMultipartUploadResult>')
          MockResponse('http://localhost:9000').put('/bucket/object?partNumber=1&uploadId=uploadid', (body) => {
            return body.length === 5 * 1024 * 1024
          }).reply(200, '', {
            etag: 'etag1'
          })
          MockResponse('http://localhost:9000').put('/bucket/object?partNumber=2&uploadId=uploadid', (body) => {
            return body.length === 5 * 1024 * 1024

          }).reply(200, '', {
            etag: 'etag2'
          })
          var s = new Stream.Readable()
          s._read = function() {}
          for (var i = 0; i < 11 * 1024; i++) {
            s.push(uploadData)
          }
          s.push(null)
          client.putObject('bucket', 'object', '', 12 * 1024 * 1024, s, (e) => {
            assert.equal(e, 'actual size does not match specified size')
            done()
          })
        })
        it('should fail if actual size is larger than expected', (done) => {
          MockResponse('http://localhost:9000').get('/bucket?uploads&max-uploads=1000&prefix=object').reply(200, '<ListMultipartUploadsResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Bucket>golang</Bucket><KeyMarker></KeyMarker><UploadIdMarker></UploadIdMarker><NextKeyMarker></NextKeyMarker><NextUploadIdMarker></NextUploadIdMarker><EncodingType></EncodingType><MaxUploads>1000</MaxUploads><IsTruncated>false</IsTruncataed><Prefix></Prefix><Delimiter></Delimiter></ListMultipartUploadsResult>')
          MockResponse('http://localhost:9000').post('/bucket/object?uploads').reply(200, '<?xml version="1.0" encoding="UTF-8"?>\n<InitiateMultipartUploadResult><Bucket>bucket</Bucket><Key>object</Key><UploadId>uploadid</UploadId></InitiateMultipartUploadResult>')
          MockResponse('http://localhost:9000').put('/bucket/object?partNumber=1&uploadId=uploadid', (body) => {
            return body.length === 5 * 1024 * 1024
          }).reply(200, '', {
            etag: 'etag1'
          })
          MockResponse('http://localhost:9000').put('/bucket/object?partNumber=2&uploadId=uploadid', (body) => {
            return body.length === 5 * 1024 * 1024

          }).reply(200, '', {
            etag: 'etag2'
          })
          var s = new Stream.Readable()
          s._read = function() {}
          for (var i = 0; i < 12 * 1024; i++) {
            s.push(uploadData)
          }
          s.push(null)
          client.putObject('bucket', 'object', '', 11 * 1024 * 1024, s, (e) => {
            assert.equal(e, 'actual size does not match specified size')
            done()
          })
        })
        it('should pass upload list error to callback', (done) => {
          MockResponse('http://localhost:9000').get('/bucket?uploads&max-uploads=1000&prefix=object').reply(400, generateError('code', 'message', 'requestid', 'hostid', 'resource'))
          var s = new Stream.Readable()
          s._read = function() {}
          for (var i = 0; i < 11 * 1024; i++) {
            s.push(uploadData)
          }
          s.push(null)
          client.putObject('bucket', 'object', '', 11 * 1024 * 1024, s, checkError('code', 'message', 'requestid', 'hostid', 'resource', done))
        })
        it('should pass part list error to callback', (done) => {
          MockResponse('http://localhost:9000').get('/bucket?uploads&max-uploads=1000&prefix=object').reply(200, '<ListMultipartUploadsResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Bucket>bucket</Bucket><KeyMarker></KeyMarker><UploadIdMarker></UploadIdMarker><NextKeyMarker></NextKeyMarker><NextUploadIdMarker></NextUploadIdMarker><EncodingType></EncodingType><MaxUploads>1000</MaxUploads><IsTruncated>false</IsTruncated><Upload><Key>object</Key><UploadId>uploadid</UploadId><Initiator><ID></ID><DisplayName></DisplayName></Initiator><Owner><ID></ID><DisplayName></DisplayName></Owner><StorageClass></StorageClass><Initiated>2015-05-30T14:43:35.349Z</Initiated></Upload><Prefix>object</Prefix><Delimiter></Delimiter></ListMultipartUploadsResult>')
          MockResponse('http://localhost:9000').get('/bucket/object?uploadId=uploadid').reply(400, generateError('code', 'message', 'requestid', 'hostid', 'resource'))
          var s = new Stream.Readable()
          s._read = function() {}
          for (var i = 0; i < 11 * 1024; i++) {
            s.push(uploadData)
          }
          s.push(null)
          client.putObject('bucket', 'object', '', 11 * 1024 * 1024, s, checkError('code', 'message', 'requestid', 'hostid', 'resource', done))
        })
        it('should pass put error to callback', (done) => {
          MockResponse('http://localhost:9000').get('/bucket?uploads&max-uploads=1000&prefix=object').reply(200, '<ListMultipartUploadsResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Bucket>bucket</Bucket><KeyMarker></KeyMarker><UploadIdMarker></UploadIdMarker><NextKeyMarker></NextKeyMarker><NextUploadIdMarker></NextUploadIdMarker><EncodingType></EncodingType><MaxUploads>1000</MaxUploads><IsTruncated>false</IsTruncated><Upload><Key>object</Key><UploadId>uploadid</UploadId><Initiator><ID></ID><DisplayName></DisplayName></Initiator><Owner><ID></ID><DisplayName></DisplayName></Owner><StorageClass></StorageClass><Initiated>2015-05-30T14:43:35.349Z</Initiated></Upload><Prefix>object</Prefix><Delimiter></Delimiter></ListMultipartUploadsResult>')
          MockResponse('http://localhost:9000').get('/bucket/object?uploadId=uploadid').reply(200, '<ListPartsResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Bucket>bucket</Bucket><Key>go1.4.2</Key><UploadId>ntWSjzBytPT2xKLaMRonzXncsO10EH4Fc-Iq2-4hG-ulRYB</UploadId><Initiator><ID>minio</ID><DisplayName>minio</DisplayName></Initiator><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner><StorageClass>STANDARD</StorageClass><PartNumberMarker>0</PartNumberMarker><NextPartNumberMarker>0</NextPartNumberMarker><MaxParts>1000</MaxParts><IsTruncated>false</IsTruncated><Part><PartNumber>1</PartNumber><ETag>"79b281060d337b9b2b84ccf390adcf74"</ETag><LastModified>2015-06-03T03:12:34.756Z</LastModified><Size>5242880</Size></Part><Part><PartNumber>2</PartNumber><ETag>"79b281060d337b9b2b84ccf390adcf74"</ETag><LastModified>2015-06-03T03:12:34.756Z</LastModified><Size>5242880</Size></Part></ListPartsResult>')
          MockResponse('http://localhost:9000').put('/bucket/object?partNumber=3&uploadId=uploadid', (body) => {
            return body.length === 1024 * 1024

          }).reply(400, generateError('code', 'message', 'requestid', 'hostid', 'resource'))
          var s = new Stream.Readable()
          s._read = function() {}
          for (var i = 0; i < 11 * 1024; i++) {
            s.push(uploadData)
          }
          s.push(null)
          client.putObject('bucket', 'object', '', 11 * 1024 * 1024, s, checkError('code', 'message', 'requestid', 'hostid', 'resource', done))
        })
        it('should pass complete upload error to callback', (done) => {
          MockResponse('http://localhost:9000').get('/bucket?uploads&max-uploads=1000&prefix=object').reply(200, '<ListMultipartUploadsResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Bucket>bucket</Bucket><KeyMarker></KeyMarker><UploadIdMarker></UploadIdMarker><NextKeyMarker></NextKeyMarker><NextUploadIdMarker></NextUploadIdMarker><EncodingType></EncodingType><MaxUploads>1000</MaxUploads><IsTruncated>false</IsTruncated><Upload><Key>object</Key><UploadId>uploadid</UploadId><Initiator><ID></ID><DisplayName></DisplayName></Initiator><Owner><ID></ID><DisplayName></DisplayName></Owner><StorageClass></StorageClass><Initiated>2015-05-30T14:43:35.349Z</Initiated></Upload><Prefix>object</Prefix><Delimiter></Delimiter></ListMultipartUploadsResult>')
          MockResponse('http://localhost:9000').get('/bucket/object?uploadId=uploadid').reply(200, '<ListPartsResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Bucket>bucket</Bucket><Key>go1.4.2</Key><UploadId>ntWSjzBytPT2xKLaMRonzXncsO10EH4Fc-Iq2-4hG-ulRYB</UploadId><Initiator><ID>minio</ID><DisplayName>minio</DisplayName></Initiator><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner><StorageClass>STANDARD</StorageClass><PartNumberMarker>0</PartNumberMarker><NextPartNumberMarker>0</NextPartNumberMarker><MaxParts>1000</MaxParts><IsTruncated>false</IsTruncated><Part><PartNumber>1</PartNumber><ETag>"79b281060d337b9b2b84ccf390adcf74"</ETag><LastModified>2015-06-03T03:12:34.756Z</LastModified><Size>5242880</Size></Part><Part><PartNumber>2</PartNumber><ETag>"79b281060d337b9b2b84ccf390adcf74"</ETag><LastModified>2015-06-03T03:12:34.756Z</LastModified><Size>5242880</Size></Part></ListPartsResult>')
          MockResponse('http://localhost:9000').put('/bucket/object?partNumber=3&uploadId=uploadid', (body) => {
            return body.length === 1024 * 1024

          }).reply(200, '', {
            etag: '79b281060d337b9b2b84ccf390adcf74'
          })
          MockResponse('http://localhost:9000').post('/bucket/object?uploadId=uploadid').reply(400, generateError('code', 'message', 'requestid', 'hostid', 'resource'))
          var s = new Stream.Readable()
          s._read = function() {}
          for (var i = 0; i < 11 * 1024; i++) {
            s.push(uploadData)
          }
          s.push(null)
          client.putObject('bucket', 'object', '', 11 * 1024 * 1024, s, checkError('code', 'message', 'requestid', 'hostid', 'resource', done))
        })
      })
    })

    describe('#listObjects()', () => {
      it('should iterate without a prefix', (done) => {
        MockResponse('http://localhost:9000').get('/bucket?max-keys=1000').reply(200, '<ListBucketResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Name>bucket</Name><Prefix></Prefix><Marker></Marker><MaxKeys>1000</MaxKeys><Delimiter></Delimiter><IsTruncated>true</IsTruncated><Contents><Key>key1</Key><LastModified>2015-05-05T02:21:15.716Z</LastModified><ETag>"5eb63bbbe01eeed093cb22bb8f5acdc3"</ETag><Size>11</Size><StorageClass>STANDARD</StorageClass><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner></Contents><Contents><Key>key2</Key><LastModified>2015-05-05T20:36:17.498Z</LastModified><ETag>"2a60eaffa7a82804bdc682ce1df6c2d4"</ETag><Size>1661</Size><StorageClass>STANDARD</StorageClass><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner></Contents></ListBucketResult>')
        MockResponse('http://localhost:9000').get('/bucket?marker=key2&max-keys=1000').reply(200, '<ListBucketResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Name>bucket</Name><Prefix></Prefix><Marker></Marker><MaxKeys>1000</MaxKeys><Delimiter></Delimiter><IsTruncated>true</IsTruncated><Contents><Key>key3</Key><LastModified>2015-05-05T02:21:15.716Z</LastModified><ETag>"5eb63bbbe01eeed093cb22bb8f5acdc3"</ETag><Size>11</Size><StorageClass>STANDARD</StorageClass><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner></Contents><Contents><Key>key4</Key><LastModified>2015-05-05T20:36:17.498Z</LastModified><ETag>"2a60eaffa7a82804bdc682ce1df6c2d4"</ETag><Size>1661</Size><StorageClass>STANDARD</StorageClass><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner></Contents></ListBucketResult>')
        MockResponse('http://localhost:9000').get('/bucket?marker=key4&max-keys=1000').reply(200, '<ListBucketResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Name>bucket</Name><Prefix></Prefix><Marker></Marker><MaxKeys>1000</MaxKeys><Delimiter></Delimiter><IsTruncated>false</IsTruncated><Contents><Key>key5</Key><LastModified>2015-05-05T02:21:15.716Z</LastModified><ETag>"5eb63bbbe01eeed093cb22bb8f5acdc3"</ETag><Size>11</Size><StorageClass>STANDARD</StorageClass><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner></Contents><Contents><Key>key6</Key><LastModified>2015-05-05T20:36:17.498Z</LastModified><ETag>"2a60eaffa7a82804bdc682ce1df6c2d4"</ETag><Size>1661</Size><StorageClass>STANDARD</StorageClass><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner></Contents></ListBucketResult>')
        var stream = client.listObjects('bucket', '', true),
          results = [],
          expectedResults = [{
            'etag': '5eb63bbbe01eeed093cb22bb8f5acdc3',
            'lastModified': '2015-05-05T02:21:15.716Z',
            'name': 'key1',
            'size': 11
          }, {
            'etag': '2a60eaffa7a82804bdc682ce1df6c2d4',
            'lastModified': '2015-05-05T20:36:17.498Z',
            'name': 'key2',
            'size': 1661
          }, {
            'etag': '5eb63bbbe01eeed093cb22bb8f5acdc3',
            'lastModified': '2015-05-05T02:21:15.716Z',
            'name': 'key3',
            'size': 11
          }, {
            'etag': '2a60eaffa7a82804bdc682ce1df6c2d4',
            'lastModified': '2015-05-05T20:36:17.498Z',
            'name': 'key4',
            'size': 1661
          }, {
            'etag': '5eb63bbbe01eeed093cb22bb8f5acdc3',
            'lastModified': '2015-05-05T02:21:15.716Z',
            'name': 'key5',
            'size': 11
          }, {
            'etag': '2a60eaffa7a82804bdc682ce1df6c2d4',
            'lastModified': '2015-05-05T20:36:17.498Z',
            'name': 'key6',
            'size': 1661
          }]
        stream.pipe(Through2.obj(function(object, enc, end) {
          results.push(object)
          end()
        }, function(end) {
          assert.deepEqual(results, expectedResults)
          end()
          done()
        }))
      })
      it('should iterate with a prefix', (done) => {
        MockResponse('http://localhost:9000').get('/bucket?max-keys=1000&prefix=key').reply(200, '<ListBucketResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Name>bucket</Name><Prefix></Prefix><Marker></Marker><MaxKeys>1000</MaxKeys><Delimiter></Delimiter><IsTruncated>true</IsTruncated><Contents><Key>key1</Key><LastModified>2015-05-05T02:21:15.716Z</LastModified><ETag>"5eb63bbbe01eeed093cb22bb8f5acdc3"</ETag><Size>11</Size><StorageClass>STANDARD</StorageClass><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner></Contents><Contents><Key>key2</Key><LastModified>2015-05-05T20:36:17.498Z</LastModified><ETag>"2a60eaffa7a82804bdc682ce1df6c2d4"</ETag><Size>1661</Size><StorageClass>STANDARD</StorageClass><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner></Contents></ListBucketResult>')
        MockResponse('http://localhost:9000').get('/bucket?marker=key2&max-keys=1000&prefix=key').reply(200, '<ListBucketResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Name>bucket</Name><Prefix></Prefix><Marker></Marker><MaxKeys>1000</MaxKeys><Delimiter></Delimiter><IsTruncated>true</IsTruncated><Contents><Key>key3</Key><LastModified>2015-05-05T02:21:15.716Z</LastModified><ETag>"5eb63bbbe01eeed093cb22bb8f5acdc3"</ETag><Size>11</Size><StorageClass>STANDARD</StorageClass><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner></Contents><Contents><Key>key4</Key><LastModified>2015-05-05T20:36:17.498Z</LastModified><ETag>"2a60eaffa7a82804bdc682ce1df6c2d4"</ETag><Size>1661</Size><StorageClass>STANDARD</StorageClass><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner></Contents></ListBucketResult>')
        MockResponse('http://localhost:9000').get('/bucket?marker=key4&max-keys=1000&prefix=key').reply(200, '<ListBucketResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Name>bucket</Name><Prefix></Prefix><Marker></Marker><MaxKeys>1000</MaxKeys><Delimiter></Delimiter><IsTruncated>false</IsTruncated><Contents><Key>key5</Key><LastModified>2015-05-05T02:21:15.716Z</LastModified><ETag>"5eb63bbbe01eeed093cb22bb8f5acdc3"</ETag><Size>11</Size><StorageClass>STANDARD</StorageClass><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner></Contents><Contents><Key>key6</Key><LastModified>2015-05-05T20:36:17.498Z</LastModified><ETag>"2a60eaffa7a82804bdc682ce1df6c2d4"</ETag><Size>1661</Size><StorageClass>STANDARD</StorageClass><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner></Contents></ListBucketResult>')
        var stream = client.listObjects('bucket', 'key', true),
          results = [],
          expectedResults = [{
            'etag': '5eb63bbbe01eeed093cb22bb8f5acdc3',
            'lastModified': '2015-05-05T02:21:15.716Z',
            'name': 'key1',
            'size': 11
          }, {
            'etag': '2a60eaffa7a82804bdc682ce1df6c2d4',
            'lastModified': '2015-05-05T20:36:17.498Z',
            'name': 'key2',
            'size': 1661
          }, {
            'etag': '5eb63bbbe01eeed093cb22bb8f5acdc3',
            'lastModified': '2015-05-05T02:21:15.716Z',
            'name': 'key3',
            'size': 11
          }, {
            'etag': '2a60eaffa7a82804bdc682ce1df6c2d4',
            'lastModified': '2015-05-05T20:36:17.498Z',
            'name': 'key4',
            'size': 1661
          }, {
            'etag': '5eb63bbbe01eeed093cb22bb8f5acdc3',
            'lastModified': '2015-05-05T02:21:15.716Z',
            'name': 'key5',
            'size': 11
          }, {
            'etag': '2a60eaffa7a82804bdc682ce1df6c2d4',
            'lastModified': '2015-05-05T20:36:17.498Z',
            'name': 'key6',
            'size': 1661
          }]
        stream.pipe(Through2.obj(function(object, enc, end) {
          results.push(object)
          end()
        }, function(end) {
          assert.deepEqual(results, expectedResults)
          end()
          done()
        }))
      })
      it.skip('should iterate with recursion', (done) => {
        MockResponse('http://localhost:9000').get('/bucket?max-keys=1000').reply(200, '<ListBucketResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Name>bucket</Name><Prefix></Prefix><Marker></Marker><MaxKeys>1000</MaxKeys><Delimiter></Delimiter><IsTruncated>true</IsTruncated><Contents><Key>key1</Key><LastModified>2015-05-05T02:21:15.716Z</LastModified><ETag>"5eb63bbbe01eeed093cb22bb8f5acdc3"</ETag><Size>11</Size><StorageClass>STANDARD</StorageClass><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner></Contents><Contents><Key>key2</Key><LastModified>2015-05-05T20:36:17.498Z</LastModified><ETag>"2a60eaffa7a82804bdc682ce1df6c2d4"</ETag><Size>1661</Size><StorageClass>STANDARD</StorageClass><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner></Contents></ListBucketResult>')
        MockResponse('http://localhost:9000').get('/bucket?marker=key2&max-keys=1000').reply(200, '<ListBucketResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Name>bucket</Name><Prefix></Prefix><Marker></Marker><MaxKeys>1000</MaxKeys><Delimiter></Delimiter><IsTruncated>true</IsTruncated><Contents><Key>key3</Key><LastModified>2015-05-05T02:21:15.716Z</LastModified><ETag>"5eb63bbbe01eeed093cb22bb8f5acdc3"</ETag><Size>11</Size><StorageClass>STANDARD</StorageClass><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner></Contents><Contents><Key>key4</Key><LastModified>2015-05-05T20:36:17.498Z</LastModified><ETag>"2a60eaffa7a82804bdc682ce1df6c2d4"</ETag><Size>1661</Size><StorageClass>STANDARD</StorageClass><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner></Contents></ListBucketResult>')
        MockResponse('http://localhost:9000').get('/bucket?marker=key4&max-keys=1000').reply(200, '<ListBucketResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Name>bucket</Name><Prefix></Prefix><Marker></Marker><MaxKeys>1000</MaxKeys><Delimiter></Delimiter><IsTruncated>false</IsTruncated><Contents><Key>key5</Key><LastModified>2015-05-05T02:21:15.716Z</LastModified><ETag>"5eb63bbbe01eeed093cb22bb8f5acdc3"</ETag><Size>11</Size><StorageClass>STANDARD</StorageClass><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner></Contents><Contents><Key>key6</Key><LastModified>2015-05-05T20:36:17.498Z</LastModified><ETag>"2a60eaffa7a82804bdc682ce1df6c2d4"</ETag><Size>1661</Size><StorageClass>STANDARD</StorageClass><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner></Contents></ListBucketResult>')
        var stream = client.listObjects('bucket', '', true),
          results = [],
          expectedResults = [{
            'etag': '5eb63bbbe01eeed093cb22bb8f5acdc3',
            'lastModified': '2015-05-05T02:21:15.716Z',
            'name': 'key1',
            'size': 11
          }, {
            'etag': '2a60eaffa7a82804bdc682ce1df6c2d4',
            'lastModified': '2015-05-05T20:36:17.498Z',
            'name': 'key2',
            'size': 1661
          }, {
            'etag': '5eb63bbbe01eeed093cb22bb8f5acdc3',
            'lastModified': '2015-05-05T02:21:15.716Z',
            'name': 'key3',
            'size': 11
          }, {
            'etag': '2a60eaffa7a82804bdc682ce1df6c2d4',
            'lastModified': '2015-05-05T20:36:17.498Z',
            'name': 'key4',
            'size': 1661
          }, {
            'etag': '5eb63bbbe01eeed093cb22bb8f5acdc3',
            'lastModified': '2015-05-05T02:21:15.716Z',
            'name': 'key5',
            'size': 11
          }, {
            'etag': '2a60eaffa7a82804bdc682ce1df6c2d4',
            'lastModified': '2015-05-05T20:36:17.498Z',
            'name': 'key6',
            'size': 1661
          }]
        stream.pipe(Through2.obj(function(object, enc, end) {
          results.push(object)
          end()
        }, function(end) {
          assert.deepEqual(results, expectedResults)
          end()
          done()
        }))
      })
      it('should pass error on stream', (done) => {
        MockResponse('http://localhost:9000').filteringPath(() => {
          return '/bucket'
        }).get('/bucket').reply(400, generateError('code', 'message', 'requestid', 'hostid', '/bucket'))
        var stream = client.listObjects('bucket')
        stream.pipe(Through2.obj(function(part, enc, end) {
          end()
        }, function(end) {
          end()
        }))
        stream.on('error', (e) => {
          checkError('code', 'message', 'requestid', 'hostid', '/bucket')(e)
          done()
        })
      })
      it('should pass error in stream on subsequent error', (done) => {
        MockResponse('http://localhost:9000').filteringPath(() => {
          return '/bucket'
        }).get('/bucket').reply(200, '<ListBucketResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Name>bucket</Name><Prefix></Prefix><Marker></Marker><MaxKeys>1000</MaxKeys><Delimiter></Delimiter><IsTruncated>true</IsTruncated><Contents><Key>key1</Key><LastModified>2015-05-05T02:21:15.716Z</LastModified><ETag>"5eb63bbbe01eeed093cb22bb8f5acdc3"</ETag><Size>11</Size><StorageClass>STANDARD</StorageClass><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner></Contents><Contents><Key>key2</Key><LastModified>2015-05-05T20:36:17.498Z</LastModified><ETag>"2a60eaffa7a82804bdc682ce1df6c2d4"</ETag><Size>1661</Size><StorageClass>STANDARD</StorageClass><Owner><ID>minio</ID><DisplayName>minio</DisplayName></Owner></Contents></ListBucketResult>')
        MockResponse('http://localhost:9000').filteringPath(() => {
          return '/bucket'
        }).get('/bucket').reply(400, generateError('code', 'message', 'requestid', 'hostid', '/bucket'))
        var stream = client.listObjects('bucket')
        stream.pipe(Through2.obj(function(part, enc, end) {
          end()
        }, function(end) {
          end()
        }))
        stream.on('error', (e) => {
          checkError('code', 'message', 'requestid', 'hostid', '/bucket')(e)
          done()
        })
      })
    })

    describe('#statObject(bucket, object, callback)', () => {
      it('should retrieve object metadata', (done) => {
        MockResponse('http://localhost:9000').head('/bucket/object').reply(200, '', {
          'ETag': 'etag',
          'Content-Length': 11,
          'Last-Modified': 'lastmodified',
          'Content-Type': 'text/plain'
        })

        client.statObject('bucket', 'object', (e, r) => {
          assert.deepEqual(r, {
            size: 11,
            lastModified: 'lastmodified',
            etag: 'etag',
            contentType: 'text/plain'
          })
          done()
        })
      })
      it('should pass error to callback', (done) => {
        MockResponse('http://localhost:9000').head('/bucket/object')
          .reply(400, generateError('code', 'message', 'requestid', 'hostid', 'resource'))

        client.statObject('bucket', 'object', checkError('code', 'message', 'requestid', 'hostid', 'resource', (r) => {
          assert.equal(r, null)
          done()
        }))
      })
      it('should fail on null bucket', (done) => {
        try {
          client.statObject(null, 'hello')
        } catch (e) {
          done()
        }
      })
      it('should fail on empty bucket', (done) => {
        try {
          client.statObject('', 'hello')
        } catch (e) {
          done()
        }
      })
      it('should fail on empty bucket', (done) => {
        try {
          client.statObject('  \n  \t  ', 'hello')
        } catch (e) {
          done()
        }
      })
      it('should fail on null object', (done) => {
        try {
          client.statObject('hello', null)
        } catch (e) {
          done()
        }
      })
      it('should fail on empty object', (done) => {
        try {
          client.statObject('hello', '')
        } catch (e) {
          done()
        }
      })
      it('should fail on empty object', (done) => {
        try {
          client.statObject('hello', '  \n  \t  ')
        } catch (e) {
          done()
        }
      })
    })

    describe('#removeObject(bucket, object, callback)', () => {
      it('should delete an object', (done) => {
        MockResponse('http://localhost:9000').delete('/bucket/object').reply(204)
        client.removeObject('bucket', 'object', (e) => {
          assert.equal(e, null)
          done()
        })
      })
      it('should pass error to callback', (done) => {
        MockResponse('http://localhost:9000').delete('/bucket/object')
          .reply(400, generateError('code', 'message', 'requestid', 'hostid', 'resource'))
        client.removeObject('bucket', 'object', checkError('code', 'message', 'requestid', 'hostid', 'resource', (r) => {
          assert.equal(r, null)
          done()
        }))
      })
      it('should fail on null bucket', (done) => {
        try {
          client.removeObject(null, 'hello')
        } catch (e) {
          done()
        }
      })
      it('should fail on empty bucket', (done) => {
        try {
          client.removeObject('', 'hello')
        } catch (e) {
          done()
        }
      })
      it('should fail on empty bucket', (done) => {
        try {
          client.removeObject('  \n  \t  ', 'hello')
        } catch (e) {
          done()
        }
      })
      it('should fail on null object', (done) => {
        try {
          client.removeObject('hello', null)
        } catch (e) {
          done()
        }
      })
      it('should fail on empty object', (done) => {
        try {
          client.removeObject('hello', '')
        } catch (e) {
          done()
        }
      })
      it('should fail on empty object', (done) => {
        try {
          client.removeObject('hello', '  \n  \t  ')
        } catch (e) {
          done()
        }
      })
    })

    describe('#removeIncompleteUpload(bucket, object, callback)', () => {
      it('should remove an incomplete upload', (done) => {
        MockResponse('http://localhost:9000').get('/golang?uploads&max-uploads=1000&prefix=go1.4.2').reply(200, '<ListMultipartUploadsResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Bucket>golang</Bucket><KeyMarker></KeyMarker><UploadIdMarker></UploadIdMarker><NextKeyMarker>keymarker</NextKeyMarker><NextUploadIdMarker>uploadmarker</NextUploadIdMarker><EncodingType></EncodingType><MaxUploads>1000</MaxUploads><IsTruncated>true</IsTruncated><Upload><Key>go1.4.2.1</Key><UploadId>uploadid</UploadId><Initiator><ID></ID><DisplayName></DisplayName></Initiator><Owner><ID></ID><DisplayName></DisplayName></Owner><StorageClass></StorageClass><Initiated>2015-05-30T14:43:35.349Z</Initiated></Upload><Upload><Key>go1.4.2</Key><UploadId>uploadid2</UploadId><Initiator><ID></ID><DisplayName></DisplayName></Initiator><Owner><ID></ID><DisplayName></DisplayName></Owner><StorageClass></StorageClass><Initiated>2015-05-30T15:00:07.759Z</Initiated></Upload><Upload><Key>go1.5.0</Key><UploadId>uploadid2</UploadId><Initiator><ID></ID><DisplayName></DisplayName></Initiator><Owner><ID></ID><DisplayName></DisplayName></Owner><StorageClass></StorageClass><Initiated>2015-05-30T15:00:07.759Z</Initiated></Upload><Prefix></Prefix><Delimiter></Delimiter></ListMultipartUploadsResult>')
        MockResponse('http://localhost:9000').delete('/golang/go1.4.2?uploadId=uploadid2').reply(204)
        client.removeIncompleteUpload('golang', 'go1.4.2', done)
      })
      it('should pass error to callback on list failure', (done) => {
        MockResponse('http://localhost:9000').get('/golang?uploads&max-uploads=1000&prefix=go1.4.2').reply(400, generateError('code', 'message', 'requestid', 'hostid', 'resource'))
        client.removeIncompleteUpload('golang', 'go1.4.2', checkError('code', 'message', 'requestid', 'hostid', 'resource', done))
      })
      it('should pass error to callback on second list failure', (done) => {
        MockResponse('http://localhost:9000').get('/golang?uploads&max-uploads=1000&prefix=go1.4.2').reply(200, '<ListMultipartUploadsResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Bucket>golang</Bucket><KeyMarker></KeyMarker><UploadIdMarker></UploadIdMarker><NextKeyMarker>keymarker</NextKeyMarker><NextUploadIdMarker>uploadmarker</NextUploadIdMarker><EncodingType></EncodingType><MaxUploads>1000</MaxUploads><IsTruncated>true</IsTruncated><Upload><Key>go1.4.2.1</Key><UploadId>uploadid</UploadId><Initiator><ID></ID><DisplayName></DisplayName></Initiator><Owner><ID></ID><DisplayName></DisplayName></Owner><StorageClass></StorageClass><Initiated>2015-05-30T14:43:35.349Z</Initiated></Upload><Upload><Key>go1.4.2.2</Key><UploadId>uploadid2</UploadId><Initiator><ID></ID><DisplayName></DisplayName></Initiator><Owner><ID></ID><DisplayName></DisplayName></Owner><StorageClass></StorageClass><Initiated>2015-05-30T15:00:07.759Z</Initiated></Upload><Prefix></Prefix><Delimiter></Delimiter></ListMultipartUploadsResult>')
        MockResponse('http://localhost:9000').get('/golang?uploads&key-marker=keymarker&max-uploads=1000&prefix=go1.4.2&upload-id-marker=uploadmarker').reply(400, generateError('code', 'message', 'requestid', 'hostid', 'resource'))
        client.removeIncompleteUpload('golang', 'go1.4.2', checkError('code', 'message', 'requestid', 'hostid', 'resource', done))
      })
      it('should return error on delete failure', (done) => {
        MockResponse('http://localhost:9000').get('/golang?uploads&max-uploads=1000&prefix=go1.4.2').reply(200, '<ListMultipartUploadsResult xmlns="http://s3.amazonaws.com/doc/2006-03-01"><Bucket>golang</Bucket><KeyMarker></KeyMarker><UploadIdMarker></UploadIdMarker><NextKeyMarker>keymarker</NextKeyMarker><NextUploadIdMarker>uploadmarker</NextUploadIdMarker><EncodingType></EncodingType><MaxUploads>1000</MaxUploads><IsTruncated>true</IsTruncated><Upload><Key>go1.4.2</Key><UploadId>uploadid</UploadId><Initiator><ID></ID><DisplayName></DisplayName></Initiator><Owner><ID></ID><DisplayName></DisplayName></Owner><StorageClass></StorageClass><Initiated>2015-05-30T14:43:35.349Z</Initiated></Upload><Upload><Key>go1.4.2.1</Key><UploadId>uploadid2</UploadId><Initiator><ID></ID><DisplayName></DisplayName></Initiator><Owner><ID></ID><DisplayName></DisplayName></Owner><StorageClass></StorageClass><Initiated>2015-05-30T15:00:07.759Z</Initiated></Upload><Prefix></Prefix><Delimiter></Delimiter></ListMultipartUploadsResult>')
        MockResponse('http://localhost:9000').delete('/golang/go1.4.2?uploadId=uploadid').reply(400, generateError('code', 'message', 'requestid', 'hostid', 'resource'))
        client.removeIncompleteUpload('golang', 'go1.4.2', checkError('code', 'message', 'requestid', 'hostid', 'resource', done))
      })
      it('should fail on null bucket', (done) => {
        try {
          client.removeIncompleteUpload(null, 'hello')
        } catch (e) {
          done()
        }
      })
      it('should fail on empty bucket', (done) => {
        try {
          client.removeIncompleteUpload('', 'hello')
        } catch (e) {
          done()
        }
      })
      it('should fail on empty bucket', (done) => {
        try {
          client.removeIncompleteUpload('  \n  \t  ', 'hello')
        } catch (e) {
          done()
        }
      })
      it('should fail on null object', (done) => {
        try {
          client.removeIncompleteUpload('hello', null)
        } catch (e) {
          done()
        }
      })
      it('should fail on empty object', (done) => {
        try {
          client.removeIncompleteUpload('hello', '')
        } catch (e) {
          done()
        }
      })
      it('should fail on empty object', (done) => {
        try {
          client.removeIncompleteUpload('hello', '  \n  \t  ')
        } catch (e) {
          done()
        }
      })
    })
  })
})

var checkError = (code, message, requestid, hostid, resource, callback) => {
  return (e, ...rest) => {
    if (e === null) {
      callback('expected error, received success')
    }
    assert.equal(e.code, code)
    assert.equal(e.message, message)
    assert.equal(e.requestid, requestid)
    assert.equal(e.hostid, hostid)
    assert.equal(e.resource, resource)
    if (callback) {
      if (rest.length === 0) {
        callback()
      } else {
        callback(rest)
      }
    } else {
      if (rest.length > 0) {
        assert.fail('Data returned with no callback registered')
      }
    }
  }
}

var generateError = (code, message, requestid, hostid, resource) => {
  return `<Error><Code>${code}</Code><Message>${message}</Message><RequestId>${requestid}</RequestId><HostId>${hostid}</HostId><Resource>${resource}</Resource></Error>`
}
